@@ -12,11 +12,14 @@ sequence:
     low level iterator classes
 sample:
     low level random sampling routines
+datatypes:
+    low level data only classes and enums
 """
 
-__version__ = "1.1.1"
+__version__ = "1.2.1"
 
 from . import config
+from . import datatypes
 from . import sample
 from . import sequence
 from . import playlist
