@@ -5,11 +5,14 @@ __all__ = ["Playlist", "ensure_state"]
 from pathlib import Path
 import pickle
 import warnings
+import time
+import os
 
-from lxml import etree
+from lxml import etree  # type:ignore
 
 import nextsong.sequence as seq
-from nextsong.config import get as get_config
+from nextsong.config import Config, get as get_cfg
+from nextsong.datatypes import OnChange
 
 
 class Playlist:
@@ -41,8 +44,9 @@ class Playlist:
             Users should instead call Playlist.__iter__ or
             Playlist.load_state to create a PlaylistState.
             """
-            self.__iterator = iterator
+            self.__iterator = seq.WithPeek(iterator)
             self.__from_path = from_path
+            self.creation_time = time.time()
 
         def __next__(self):
             return next(self.__iterator)
@@ -56,10 +60,13 @@ class Playlist:
                 The desired path to pickle to. If None, uses the value
                 of the "state_path" config.
             """
-            if filepath is None:
-                filepath = get_config("state_path")
-            with open(filepath, "wb") as file:
-                return pickle.dump(self, file)
+            with Config(state_path=filepath) as cfg:
+                with open(cfg.state_path, "wb") as file:
+                    return pickle.dump(self, file)
+
+        def peek(self):
+            """Peek at the next track without consuming it"""
+            return self.__iterator.peek()
 
         def __enter__(self):
             return self
@@ -82,11 +89,10 @@ class Playlist:
                 If False, the FileNotFoundError is allowed to
                 propagate.
             """
-            if filepath is None:
-                filepath = get_config("state_path")
             try:
-                with open(filepath, "rb") as file:
-                    return pickle.load(file)
+                with Config(state_path=filepath) as cfg:
+                    with open(cfg.state_path, "rb") as file:
+                        return pickle.load(file)
             except FileNotFoundError:
                 if handle_not_found:
                     return None
@@ -180,9 +186,13 @@ class Playlist:
         if options["loop"]:
             if options["shuffle"]:
                 if options["count"] is not None:
-                    raise ValueError("count requires loop=False or shuffle=False")
+                    raise ValueError(
+                        "count requires loop=False or shuffle=False"
+                    )
                 if options["portion"] is not None:
-                    raise ValueError("portion requires loop=False or shuffle=False")
+                    raise ValueError(
+                        "portion requires loop=False or shuffle=False"
+                    )
             else:
                 if options["recent_portion"] is not None:
                     raise ValueError("recent_portion requires shuffle=True")
@@ -190,7 +200,9 @@ class Playlist:
             if options["recent_portion"] is not None:
                 if options["shuffle"]:
                     raise ValueError("recent_portion requires loop=True")
-                raise ValueError("recent_portion requires loop=True and shuffle=True")
+                raise ValueError(
+                    "recent_portion requires loop=True and shuffle=True"
+                )
 
     @property
     def children(self):
@@ -211,7 +223,7 @@ class Playlist:
 
     @staticmethod
     def __resolve_path(path):
-        root = Path(get_config("media_root"))
+        root = Path(get_cfg("media_root"))
         resolved_path = (root / path).resolve()
         if resolved_path.exists():
             resolved_paths = [resolved_path]
@@ -220,17 +232,21 @@ class Playlist:
             resolved_paths = [p for p in resolved_paths if p.is_file()]
             if not resolved_paths:
                 warnings.warn(
-                    f'file "{resolved_path}" not found and has no matches as a glob pattern'
+                    f'file "{resolved_path}" not found and has no matches as '
+                    "a glob pattern"
                 )
 
-        if get_config("media_exts"):
+        if get_cfg("media_exts"):
             supported_paths = []
             for resolved_path in resolved_paths:
-                if resolved_path.suffix.lower().lstrip(".") in get_config("media_exts"):
+                if resolved_path.suffix.lower().lstrip(".") in get_cfg(
+                    "media_exts"
+                ):
                     supported_paths.append(resolved_path)
                 else:
                     warnings.warn(
-                        f'file "{resolved_path}" has unsupported extension and will be skipped'
+                        f'file "{resolved_path}" has unsupported extension '
+                        "and will be skipped"
                     )
         else:
             supported_paths = resolved_paths
@@ -272,6 +288,22 @@ class Playlist:
     def __iter__(self):
         return self.PlaylistState(iter(self.__create_sequence()))
 
+    def paths(self):
+        """List all filepaths that appear in the Playlist
+
+        This flattens the Playlist and any descendant Playlists and
+        resolves any glob patterns to get a flat list of all files
+        referenced by the Playlist. There are no guarantees about the
+        order or uniqueness of the listed values.
+        """
+        tracks = []
+        for child in self.children:
+            if isinstance(child, Playlist):
+                tracks.extend(child.paths())
+            elif isinstance(child, str):
+                tracks.extend(self.__resolve_path(child))
+        return tracks
+
     def save_xml(self, filepath=None):
         """Save the Playlist to an xml file
 
@@ -283,9 +315,6 @@ class Playlist:
             The desired path of the xml file. If None, uses the value of
             the "playlist_path" config.
         """
-        if filepath is None:
-            filepath = get_config("playlist_path")
-
         root = etree.Element("nextsong")
 
         meta = etree.Element("meta")
@@ -309,7 +338,9 @@ class Playlist:
                 if isinstance(val, (tuple, list)):
                     attributes[key] = " ".join(str(x) for x in val)
                     continue
-                warnings.warn(f'could not serialize option "{key}" with value "{val}"')
+                warnings.warn(
+                    f'could not serialize option "{key}" with value "{val}"'
+                )
             return attributes
 
         def to_elem(node):
@@ -329,7 +360,9 @@ class Playlist:
         root.append(elem)
 
         tree = etree.ElementTree(root)
-        tree.write(filepath, pretty_print=True)
+        with Config(playlist_path=filepath) as cfg:
+            with open(cfg.playlist_path, "wb") as file:
+                tree.write(file, pretty_print=True)
 
     @staticmethod
     def load_xml(filepath=None):
@@ -344,8 +377,6 @@ class Playlist:
             The path to the xml file. If None, uses the value of the
             "playlist_path" config.
         """
-        if filepath is None:
-            filepath = get_config("playlist_path")
 
         def to_options(attributes):
             options = {}
@@ -364,7 +395,8 @@ class Playlist:
                         parsed_tokens.append(parse_type(token))
                     except ValueError:
                         warnings.warn(
-                            f'could not deserialize attribute "{key}" with value "{val}"'
+                            f'could not deserialize attribute "{key}" with '
+                            f'value "{val}"'
                         )
                         continue
                 if len(parsed_tokens) == 1:
@@ -384,13 +416,15 @@ class Playlist:
             warnings.warn(f'unexpected tag "{elem.tag}"')
             return None
 
-        try:
-            tree = etree.parse(filepath)
-        except OSError:
-            # lxml.etree only raises a basic OSError. Try opening
-            # ourselves to trigger a more detailed error.
-            with open(filepath, "r"):
-                pass
+        with Config(playlist_path=filepath) as cfg:
+            try:
+                tree = etree.parse(cfg.playlist_path)
+            except OSError:
+                # lxml.etree only raises a basic OSError. Try opening
+                # ourselves to trigger a more detailed error.
+                with open(cfg.playlist_path, "r"):
+                    pass
+                raise
         root = tree.getroot()
         if root.tag.lower() == "playlist":
             elem = root
@@ -405,40 +439,91 @@ class Playlist:
     load_state = PlaylistState.load
 
 
-def ensure_state(*, state_path=None, playlist_path=None, new_state=None):
+def _handle_playlist_change(state):
+    with Config() as cfg:
+        if cfg.on_change == OnChange.IGNORE:
+            state.creation_time = time.time()
+        elif cfg.on_change == OnChange.RESTART:
+            playlist = Playlist.load_xml()
+            state = iter(playlist)
+            state.save()
+        elif cfg.on_change == OnChange.SEEK:
+            playlist = Playlist.load_xml()
+            try:
+                next_track = next(state)
+            except StopIteration:
+                next_track = None
+            state = iter(playlist)
+            if next_track in playlist.paths():
+                # We've confirmed the next track should appear in the
+                # new playlist, but it's possible the track is very
+                # unlikely or even impossible to actually be reached,
+                # so there is a cap on the number of skips to avoid
+                # an infinite loop.
+                for _ in range(cfg.max_seek_skips):
+                    if state.peek() == next_track:
+                        break
+                    try:
+                        next(state)
+                    except StopIteration:
+                        state = iter(playlist)
+                        break
+                else:
+                    warnings.warn(
+                        "Gave up seeking to next track in new "
+                        f"playlist after {cfg.max_seek_skips} attempts"
+                    )
+                    state = iter(playlist)
+            else:
+                warnings.warn("Next track not in new playlist. Starting over.")
+        else:
+            raise NotImplementedError
+    return state
+
+
+def ensure_state(
+    *, state_path=None, playlist_path=None, new_state=None, on_change=None
+):
     """Loads a Playlist.PlaylistState or creates a new one
 
     This function will either load an existing PlaylistState from the
     filesystem, or if it doesn't exist, load a Playlist from the
     filesystem and use it to create a new PlaylistState.
 
-    Arguments
-    ---------
-    state_path: str or None
-        The path to the PlaylistState pickle file. If None, uses the
-        value of the "state_path" config.
-    playlist_path: str or None
-        The path to the playlist xml file. This is used to create the
-        PlaylistState if the pickle file doesn't already exist. If None,
-        uses the value of the "playlist_path" config.
-    new_state: bool or None
-        If True, forces the creation of a new PlaylistState from a
-        Playlist. If None, uses the value of the "new_state" config.
+    Keyword arguments may be set to override config values of the same
+    name. See the nextsong.config docs for more information.
     """
-    if new_state is None:
-        new_state = get_config("new_state")
+    with Config(
+        state_path=state_path,
+        playlist_path=playlist_path,
+        new_state=new_state,
+        on_change=on_change,
+    ) as cfg:
+        state = None
 
-    state = None
+        if not cfg.new_state:
+            try:
+                state = Playlist.load_state()
+            except FileNotFoundError:
+                pass
 
-    if not new_state:
-        try:
-            return Playlist.load_state(state_path)
-        except FileNotFoundError:
-            pass
+        if state:
+            playlist_mtime = -1
+            try:
+                playlist_mtime = os.path.getmtime(cfg.playlist_path)
+            except OSError as err:
+                lines = [
+                    "Failed to fetch playlist modified time with error:",
+                    f"\t{err}",
+                    "Change detection will be skipped",
+                ]
+                warnings.warn("\n".join(lines))
+            if state.creation_time < playlist_mtime:
+                state = _handle_playlist_change(state)
+                state.save()
+        else:
+            playlist = Playlist.load_xml()
+            state = iter(playlist)
+            state.save()
 
-    if state is None:
-        playlist = Playlist.load_xml(playlist_path)
-        state = iter(playlist)
-        state.save(state_path)
-
-    return state
+        return state
